/*
 *    Copyright (C) 2011 University of Southern California and
 *                  Song Qiang
 *
 *    Authors: Song Qiang
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>


#include <cassert>
#include <cmath>
#include <gsl/gsl_cdf.h>
#include <vector>

#include "nonparametric-test.hpp"
#include "numerical_utils.hpp"



using std::cerr;
using std::endl;
using std::vector;
using std::string;


static const double episilon = 1e-20;

double 
NonParametricTest::sign_test(const std::vector<double> &x,
                             const std::vector<double> &y,
                             const bool alternative)
{
    static const double probability_null = 0.5;
    assert(x.size() == y.size());

    size_t total = 0;
    size_t greater = 0;
    for (size_t i = 0; i < x.size(); ++i)
        if (fabs(x[i] - y[i]) > episilon)
        {
            ++total;
            greater += static_cast<size_t>(x[i] > y[i]);
        }

    if (total == 0)
        return 1.0;
    else if (!alternative)
        return gsl_cdf_binomial_Q(greater > 0 ? (greater - 1) : greater,
                                  probability_null,
                                  total);
    else
        return gsl_cdf_binomial_Q(
            total != greater ? (total - greater - 1) : (total - greater),
            probability_null,
            total);
}

static bool
cmp_abs(const double &a, const double &b)
{
    return fabs(a) < fabs(b);
}

// this function is defined in the end of the file
// Press Ctrl-S to search wilcoxon_p_value
static double
wilcoxon_p_value(const double &w_pos,
                 const double &w_neg,
                 const size_t &sample_size,
                 const bool &alternative);

double 
NonParametricTest::wilcoxon_test(const std::vector<double> &x,
                                 const std::vector<double> &y,
                                 const bool alternative)
{
    assert(x.size() == y.size());

    vector<double> diffs;
    for (size_t i = 0; i < x.size(); ++i)
    {
        const double diff = x[i] - y[i];
        if (fabs(diff) > episilon) diffs.push_back(diff);
    }

    std::sort(diffs.begin(), diffs.end(), cmp_abs);
    vector<double> ranks(diffs.size());
    for (size_t i = 0; i < diffs.size(); ++i) ranks[i] = i + 1;  
    size_t start = 0;
    while (start < ranks.size())
    {
        double rank_sum = ranks[start];
        size_t end = start + 1;
        while (end < ranks.size()
               && fabs(fabs(diffs[end]) - fabs(diffs[start])) < episilon)
        {
            rank_sum += ranks[end];
            ++end;
        }
        for (size_t i = start; i < end; ++i)
            ranks[i] = rank_sum / (end - start);
        start = end;
    }

    double w_pos(0), w_neg(0);
    for (size_t i = 0; i < ranks.size(); ++i)
        if (diffs[i] > 0)
            w_pos += ranks[i];
        else
            w_neg += ranks[i];

    if (ranks.size() > 0)
        return wilcoxon_p_value(w_pos, w_neg, ranks.size(), alternative);
    else
        return 1.0;
}

/***** Compute the Critical value table for given sample size ****/
// static void 
// wilcoxon_critical_value_table(const size_t &sample_size)
// {
//     vector<size_t> ranks(sample_size);
//     for (size_t i = 0; i < sample_size; ++i)
//         ranks[i] = i + 1;

//     const int max_w = (sample_size + 1) * sample_size / 2;
//     vector<size_t> frequencies(max_w + 1); // allocate extra cell for 0 
    
//     const size_t total = 1 << sample_size;
//     size_t sign = 0;
//     for (size_t i = 0; i < total; ++i)
//     {
//         size_t w_pos(0);
//         size_t mask = 1;
//         for (size_t j = 0; j < sample_size; ++j)
//         {
//             if (sign & mask) w_pos += ranks[j];
//             mask = mask << 1;
//         }

//         ++frequencies[w_pos];
//         ++sign;
//     }

//     cout << sample_size << "\t" << frequencies.size() << endl;
//     size_t cdf = 0;
//     for (size_t i = 0; i < frequencies.size(); ++i)
//     {
//         cdf += frequencies[i];
//         cout << i << "\t"
//              << frequencies[i] << "\t"
//              << cdf << endl;
//     }
// }

static const size_t w_statistic_cdf_1[] = {
1, 2};
static const size_t w_statistic_cdf_2[] = {
1, 2, 3, 4};
static const size_t w_statistic_cdf_3[] = {
1, 2, 3, 5, 6, 7, 8};
static const size_t w_statistic_cdf_4[] = {
1, 2, 3, 5, 7, 9, 11, 13, 14, 15, 16};
static const size_t w_statistic_cdf_5[] = {
1, 2, 3, 5, 7, 10, 13, 16, 19, 22, 25, 27, 29, 30, 31, 32};
static const size_t w_statistic_cdf_6[] = {
1, 2, 3, 5, 7, 10, 14, 18, 22, 27, 32, 37, 42, 46, 50, 54, 57, 59, 61, 62, 63, 64};
static const size_t w_statistic_cdf_7[] = {
1, 2, 3, 5, 7, 10, 14, 19, 24, 30, 37, 44, 52, 60, 68, 76, 84, 91, 98, 104, 109, 114, 118, 121, 123, 125, 126, 127, 128};
static const size_t w_statistic_cdf_8[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 32, 40, 49, 59, 70, 82, 95, 108, 121, 135, 148, 161, 174, 186, 197, 207, 216, 224, 231, 237, 242, 246, 249, 251, 253, 254, 255, 256};
static const size_t w_statistic_cdf_9[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 42, 52, 64, 77, 92, 109, 127, 146, 167, 188, 210, 233, 256, 279, 302, 324, 345, 366, 385, 403, 420, 435, 448, 460, 470, 479, 487, 493, 498, 502, 505, 507, 509, 510, 511, 512};
static const size_t w_statistic_cdf_10[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 54, 67, 82, 99, 119, 141, 165, 192, 221, 252, 285, 320, 356, 394, 433, 472, 512, 552, 591, 630, 668, 704, 739, 772, 803, 832, 859, 883, 905, 925, 942, 957, 970, 981, 991, 999, 1005, 1010, 1014, 1017, 1019, 1021, 1022, 1023, 1024};
static const size_t w_statistic_cdf_11[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 69, 85, 104, 126, 151, 179, 211, 246, 285, 328, 374, 423, 476, 532, 591, 653, 717, 783, 851, 920, 989, 1059, 1128, 1197, 1265, 1331, 1395, 1457, 1516, 1572, 1625, 1674, 1720, 1763, 1802, 1837, 1869, 1897, 1922, 1944, 1963, 1979, 1993, 2005, 2015, 2023, 2029, 2034, 2038, 2041, 2043, 2045, 2046, 2047, 2048};
static const size_t w_statistic_cdf_12[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 87, 107, 131, 158, 189, 225, 265, 310, 361, 417, 478, 545, 617, 695, 779, 868, 962, 1062, 1166, 1274, 1387, 1502, 1620, 1741, 1863, 1986, 2110, 2233, 2355, 2476, 2594, 2709, 2822, 2930, 3034, 3134, 3228, 3317, 3401, 3479, 3551, 3618, 3679, 3735, 3786, 3831, 3871, 3907, 3938, 3965, 3989, 4009, 4026, 4041, 4053, 4063, 4071, 4077, 4082, 4086, 4089, 4091, 4093, 4094, 4095, 4096};
static const size_t w_statistic_cdf_13[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 109, 134, 163, 196, 235, 279, 329, 386, 450, 521, 600, 687, 782, 886, 999, 1120, 1251, 1391, 1539, 1697, 1863, 2037, 2219, 2408, 2603, 2805, 3012, 3223, 3438, 3656, 3875, 4096, 4317, 4536, 4754, 4969, 5180, 5387, 5589, 5784, 5973, 6155, 6329, 6495, 6653, 6801, 6941, 7072, 7193, 7306, 7410, 7505, 7592, 7671, 7742, 7806, 7863, 7913, 7957, 7996, 8029, 8058, 8083, 8104, 8122, 8137, 8149, 8159, 8167, 8173, 8178, 8182, 8185, 8187, 8189, 8190, 8191, 8192};
static const size_t w_statistic_cdf_14[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 136, 166, 201, 242, 289, 343, 405, 475, 554, 643, 742, 852, 974, 1108, 1254, 1414, 1587, 1774, 1976, 2192, 2423, 2669, 2929, 3203, 3492, 3794, 4109, 4437, 4776, 5126, 5487, 5856, 6233, 6617, 7006, 7399, 7795, 8192, 8589, 8985, 9378, 9767, 10151, 10528, 10897, 11258, 11608, 11947, 12275, 12590, 12892, 13181, 13455, 13715, 13961, 14192, 14408, 14610, 14797, 14970, 15130, 15276, 15410, 15532, 15642, 15741, 15830, 15909, 15979, 16041, 16095, 16142, 16183, 16218, 16248, 16274, 16296, 16314, 16329, 16341, 16351, 16359, 16365, 16370, 16374, 16377, 16379, 16381, 16382, 16383, 16384};
static const size_t w_statistic_cdf_15[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 168, 204, 247, 296, 353, 419, 494, 579, 676, 785, 907, 1044, 1196, 1364, 1550, 1753, 1975, 2218, 2481, 2766, 3074, 3404, 3757, 4135, 4536, 4961, 5411, 5884, 6380, 6901, 7443, 8007, 8593, 9198, 9822, 10464, 11121, 11792, 12477, 13172, 13876, 14588, 15304, 16023, 16745, 17464, 18180, 18892, 19596, 20291, 20976, 21647, 22304, 22946, 23570, 24175, 24761, 25325, 25867, 26388, 26884, 27357, 27807, 28232, 28633, 29011, 29364, 29694, 30002, 30287, 30550, 30793, 31015, 31218, 31404, 31572, 31724, 31861, 31983, 32092, 32189, 32274, 32349, 32415, 32472, 32521, 32564, 32600, 32631, 32658, 32680, 32698, 32713, 32725, 32735, 32743, 32749, 32754, 32758, 32761, 32763, 32765, 32766, 32767, 32768};
static const size_t w_statistic_cdf_16[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 206, 250, 301, 360, 429, 508, 598, 701, 818, 950, 1099, 1266, 1452, 1660, 1890, 2143, 2422, 2728, 3062, 3427, 3823, 4251, 4714, 5212, 5746, 6318, 6928, 7576, 8265, 8993, 9760, 10568, 11416, 12303, 13230, 14195, 15196, 16234, 17307, 18412, 19549, 20715, 21907, 23125, 24365, 25623, 26899, 28189, 29489, 30798, 32111, 33425, 34738, 36047, 37347, 38637, 39913, 41171, 42411, 43629, 44821, 45987, 47124, 48229, 49302, 50340, 51341, 52306, 53233, 54120, 54968, 55776, 56543, 57271, 57960, 58608, 59218, 59790, 60324, 60822, 61285, 61713, 62109, 62474, 62808, 63114, 63393, 63646, 63876, 64084, 64270, 64437, 64586, 64718, 64835, 64938, 65028, 65107, 65176, 65235, 65286, 65330, 65367, 65399, 65426, 65448, 65466, 65481, 65493, 65503, 65511, 65517, 65522, 65526, 65529, 65531, 65533, 65534, 65535, 65536};
static const size_t w_statistic_cdf_17[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 252, 304, 365, 436, 518, 612, 720, 843, 983, 1142, 1321, 1522, 1748, 2000, 2280, 2591, 2934, 3312, 3728, 4183, 4680, 5222, 5810, 6447, 7136, 7878, 8675, 9531, 10445, 11420, 12458, 13559, 14725, 15958, 17257, 18623, 20057, 21558, 23126, 24761, 26461, 28225, 30053, 31941, 33888, 35892, 37949, 40057, 42214, 44414, 46655, 48933, 51243, 53581, 55944, 58325, 60720, 63126, 65536, 67946, 70352, 72747, 75128, 77491, 79829, 82139, 84417, 86658, 88858, 91015, 93123, 95180, 97184, 99131, 101019, 102847, 104611, 106311, 107946, 109514, 111015, 112449, 113815, 115114, 116347, 117513, 118614, 119652, 120627, 121541, 122397, 123194, 123936, 124625, 125262, 125850, 126392, 126889, 127344, 127760, 128138, 128481, 128792, 129072, 129324, 129550, 129751, 129930, 130089, 130229, 130352, 130460, 130554, 130636, 130707, 130768, 130820, 130865, 130903, 130935, 130962, 130984, 131002, 131017, 131029, 131039, 131047, 131053, 131058, 131062, 131065, 131067, 131069, 131070, 131071, 131072};
static const size_t w_statistic_cdf_18[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 306, 368, 441, 525, 622, 734, 862, 1008, 1175, 1364, 1577, 1818, 2088, 2390, 2728, 3103, 3519, 3980, 4487, 5045, 5658, 6328, 7059, 7856, 8721, 9658, 10673, 11766, 12942, 14206, 15559, 17005, 18549, 20191, 21935, 23785, 25741, 27806, 29983, 32271, 34672, 37189, 39819, 42563, 45423, 48394, 51477, 54672, 57973, 61380, 64891, 68500, 72204, 76001, 79883, 83846, 87887, 91997, 96171, 100405, 104688, 109016, 113383, 117778, 122196, 126631, 131072, 135513, 139948, 144366, 148761, 153128, 157456, 161739, 165973, 170147, 174257, 178298, 182261, 186143, 189940, 193644, 197253, 200764, 204171, 207472, 210667, 213750, 216721, 219581, 222325, 224955, 227472, 229873, 232161, 234338, 236403, 238359, 240209, 241953, 243595, 245139, 246585, 247938, 249202, 250378, 251471, 252486, 253423, 254288, 255085, 255816, 256486, 257099, 257657, 258164, 258625, 259041, 259416, 259754, 260056, 260326, 260567, 260780, 260969, 261136, 261282, 261410, 261522, 261619, 261703, 261776, 261838, 261891, 261937, 261975, 262007, 262034, 262056, 262074, 262089, 262101, 262111, 262119, 262125, 262130, 262134, 262137, 262139, 262141, 262142, 262143, 262144};
static const size_t w_statistic_cdf_19[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 370, 444, 530, 629, 744, 876, 1027, 1200, 1397, 1620, 1873, 2158, 2478, 2838, 3240, 3688, 4187, 4740, 5351, 6026, 6769, 7584, 8478, 9455, 10520, 11681, 12941, 14306, 15783, 17377, 19093, 20939, 22919, 25038, 27304, 29721, 32293, 35028, 37929, 41000, 44248, 47675, 51284, 55081, 59067, 63243, 67614, 72179, 76939, 81896, 87049, 92395, 97936, 103668, 109587, 115693, 121980, 128442, 135077, 141877, 148835, 155946, 163201, 170590, 178108, 185744, 193486, 201328, 209257, 217261, 225332, 233457, 241622, 249819, 258034, 266254, 274469, 282666, 290831, 298956, 307027, 315031, 322960, 330802, 338544, 346180, 353698, 361087, 368342, 375453, 382411, 389211, 395846, 402308, 408595, 414701, 420620, 426352, 431893, 437239, 442392, 447349, 452109, 456674, 461045, 465221, 469207, 473004, 476613, 480040, 483288, 486359, 489260, 491995, 494567, 496984, 499250, 501369, 503349, 505195, 506911, 508505, 509982, 511347, 512607, 513768, 514833, 515810, 516704, 517519, 518262, 518937, 519548, 520101, 520600, 521048, 521450, 521810, 522130, 522415, 522668, 522891, 523088, 523261, 523412, 523544, 523659, 523758, 523844, 523918, 523981, 524035, 524081, 524119, 524151, 524178, 524200, 524218, 524233, 524245, 524255, 524263, 524269, 524274, 524278, 524281, 524283, 524285, 524286, 524287, 524288};
static const size_t w_statistic_cdf_20[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 446, 533, 634, 751, 886, 1041, 1219, 1422, 1653, 1916, 2213, 2548, 2926, 3350, 3825, 4356, 4947, 5604, 6333, 7139, 8028, 9008, 10084, 11264, 12557, 13968, 15506, 17180, 18997, 20966, 23097, 25397, 27876, 30544, 33409, 36480, 39768, 43280, 47026, 51017, 55259, 59762, 64536, 69587, 74924, 80555, 86485, 92722, 99273, 106142, 113334, 120855, 128706, 136891, 145414, 154273, 163470, 173006, 182877, 193083, 203621, 214485, 225671, 237175, 248987, 261100, 273507, 286196, 299157, 312381, 325852, 339558, 353487, 367621, 381947, 396449, 411108, 425908, 440833, 455862, 470977, 486161, 501392, 516652, 531924, 547184, 562415, 577599, 592714, 607743, 622668, 637468, 652127, 666629, 680955, 695089, 709018, 722724, 736195, 749419, 762380, 775069, 787476, 799589, 811401, 822905, 834091, 844955, 855493, 865699, 875570, 885106, 894303, 903162, 911685, 919870, 927721, 935242, 942434, 949303, 955854, 962091, 968021, 973652, 978989, 984040, 988814, 993317, 997559, 1001550, 1005296, 1008808, 1012096, 1015167, 1018032, 1020700, 1023179, 1025479, 1027610, 1029579, 1031396, 1033070, 1034608, 1036019, 1037312, 1038492, 1039568, 1040548, 1041437, 1042243, 1042972, 1043629, 1044220, 1044751, 1045226, 1045650, 1046028, 1046363, 1046660, 1046923, 1047154, 1047357, 1047535, 1047690, 1047825, 1047942, 1048043, 1048130, 1048205, 1048269, 1048323, 1048369, 1048407, 1048439, 1048466, 1048488, 1048506, 1048521, 1048533, 1048543, 1048551, 1048557, 1048562, 1048566, 1048569, 1048571, 1048573, 1048574, 1048575, 1048576};
static const size_t w_statistic_cdf_21[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 535, 637, 756, 893, 1051, 1233, 1441, 1678, 1949, 2256, 2603, 2996, 3438, 3935, 4493, 5116, 5811, 6586, 7446, 8399, 9454, 10617, 11898, 13308, 14854, 16547, 18399, 20419, 22619, 25013, 27610, 30424, 33470, 36759, 40305, 44124, 48227, 52630, 57350, 62398, 67790, 73544, 79671, 86188, 93112, 100453, 108228, 116453, 125139, 134300, 143952, 154103, 164767, 175958, 187682, 199950, 212774, 226157, 240109, 254638, 269744, 285433, 301711, 318574, 336024, 354062, 372681, 391879, 411654, 431994, 452892, 474342, 496327, 518838, 541863, 565381, 589378, 613839, 638739, 664060, 689782, 715877, 742323, 769099, 796171, 823515, 851106, 878910, 906900, 935049, 963320, 991685, 1020116, 1048576, 1077036, 1105467, 1133832, 1162103, 1190252, 1218242, 1246046, 1273637, 1300981, 1328053, 1354829, 1381275, 1407370, 1433092, 1458413, 1483313, 1507774, 1531771, 1555289, 1578314, 1600825, 1622810, 1644260, 1665158, 1685498, 1705273, 1724471, 1743090, 1761128, 1778578, 1795441, 1811719, 1827408, 1842514, 1857043, 1870995, 1884378, 1897202, 1909470, 1921194, 1932385, 1943049, 1953200, 1962852, 1972013, 1980699, 1988924, 1996699, 2004040, 2010964, 2017481, 2023608, 2029362, 2034754, 2039802, 2044522, 2048925, 2053028, 2056847, 2060393, 2063682, 2066728, 2069542, 2072139, 2074533, 2076733, 2078753, 2080605, 2082298, 2083844, 2085254, 2086535, 2087698, 2088753, 2089706, 2090566, 2091341, 2092036, 2092659, 2093217, 2093714, 2094156, 2094549, 2094896, 2095203, 2095474, 2095711, 2095919, 2096101, 2096259, 2096396, 2096515, 2096617, 2096705, 2096781, 2096845, 2096899, 2096945, 2096983, 2097015, 2097042, 2097064, 2097082, 2097097, 2097109, 2097119, 2097127, 2097133, 2097138, 2097142, 2097145, 2097147, 2097149, 2097150, 2097151, 2097152};
static const size_t w_statistic_cdf_22[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 639, 759, 898, 1058, 1243, 1455, 1697, 1974, 2289, 2646, 3051, 3508, 4023, 4603, 5253, 5980, 6793, 7699, 8706, 9825, 11064, 12433, 13945, 15610, 17440, 19450, 21652, 24060, 26691, 29559, 32680, 36073, 39755, 43743, 48059, 52720, 57746, 63161, 68984, 75236, 81943, 89125, 96805, 105010, 113761, 123082, 133000, 143538, 154719, 166571, 179116, 192377, 206382, 221152, 236709, 253079, 270281, 288336, 307268, 327094, 347831, 369501, 392118, 415695, 440250, 465793, 492332, 519882, 548447, 578031, 608642, 640279, 672941, 706630, 741339, 777060, 813789, 851513, 890217, 929891, 970515, 1012067, 1054532, 1097882, 1142089, 1187130, 1232972, 1279581, 1326928, 1374974, 1423679, 1473008, 1522918, 1573363, 1624305, 1675695, 1727484, 1779630, 1832081, 1884785, 1937697, 1990763, 2043930, 2097152, 2150374, 2203541, 2256607, 2309519, 2362223, 2414674, 2466820, 2518609, 2569999, 2620941, 2671386, 2721296, 2770625, 2819330, 2867376, 2914723, 2961332, 3007174, 3052215, 3096422, 3139772, 3182237, 3223789, 3264413, 3304087, 3342791, 3380515, 3417244, 3452965, 3487674, 3521363, 3554025, 3585662, 3616273, 3645857, 3674422, 3701972, 3728511, 3754054, 3778609, 3802186, 3824803, 3846473, 3867210, 3887036, 3905968, 3924023, 3941225, 3957595, 3973152, 3987922, 4001927, 4015188, 4027733, 4039585, 4050766, 4061304, 4071222, 4080543, 4089294, 4097499, 4105179, 4112361, 4119068, 4125320, 4131143, 4136558, 4141584, 4146245, 4150561, 4154549, 4158231, 4161624, 4164745, 4167613, 4170244, 4172652, 4174854, 4176864, 4178694, 4180359, 4181871, 4183240, 4184479, 4185598, 4186605, 4187511, 4188324, 4189051, 4189701, 4190281, 4190796, 4191253, 4191658, 4192015, 4192330, 4192607, 4192849, 4193061, 4193246, 4193406, 4193545, 4193665, 4193768, 4193857, 4193933, 4193997, 4194051, 4194097, 4194135, 4194167, 4194194, 4194216, 4194234, 4194249, 4194261, 4194271, 4194279, 4194285, 4194290, 4194294, 4194297, 4194299, 4194301, 4194302, 4194303, 4194304};
static const size_t w_statistic_cdf_23[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 761, 901, 1063, 1250, 1465, 1711, 1993, 2314, 2679, 3094, 3563, 4093, 4691, 5363, 6117, 6962, 7906, 8959, 10132, 11435, 12880, 14481, 16249, 18199, 20348, 22710, 25303, 28146, 31256, 34654, 38362, 42401, 46794, 51567, 56743, 62349, 68414, 74964, 82029, 89642, 97831, 106630, 116074, 126194, 137027, 148610, 160978, 174169, 188223, 203176, 219068, 235941, 253832, 272782, 292834, 314024, 336395, 359988, 384840, 410992, 438485, 467354, 497638, 529375, 562598, 597342, 633643, 671529, 711031, 752180, 794998, 839512, 885746, 933716, 983442, 1034941, 1088222, 1143296, 1200172, 1258851, 1319335, 1381626, 1445713, 1511590, 1579248, 1648667, 1719831, 1792721, 1867306, 1943561, 2021455, 2100949, 2182005, 2264584, 2348636, 2434114, 2520969, 2609141, 2698574, 2789210, 2880980, 2973821, 3067667, 3162441, 3258073, 3354489, 3451608, 3549353, 3647646, 3746401, 3845537, 3944973, 4044620, 4144394, 4244214, 4343988, 4443635, 4543071, 4642207, 4740962, 4839255, 4937000, 5034119, 5130535, 5226167, 5320941, 5414787, 5507628, 5599398, 5690034, 5779467, 5867639, 5954494, 6039972, 6124024, 6206603, 6287659, 6367153, 6445047, 6521302, 6595887, 6668777, 6739941, 6809360, 6877018, 6942895, 7006982, 7069273, 7129757, 7188436, 7245312, 7300386, 7353667, 7405166, 7454892, 7502862, 7549096, 7593610, 7636428, 7677577, 7717079, 7754965, 7791266, 7826010, 7859233, 7890970, 7921254, 7950123, 7977616, 8003768, 8028620, 8052213, 8074584, 8095774, 8115826, 8134776, 8152667, 8169540, 8185432, 8200385, 8214439, 8227630, 8239998, 8251581, 8262414, 8272534, 8281978, 8290777, 8298966, 8306579, 8313644, 8320194, 8326259, 8331865, 8337041, 8341814, 8346207, 8350246, 8353954, 8357352, 8360462, 8363305, 8365898, 8368260, 8370409, 8372359, 8374127, 8375728, 8377173, 8378476, 8379649, 8380702, 8381646, 8382491, 8383245, 8383917, 8384515, 8385045, 8385514, 8385929, 8386294, 8386615, 8386897, 8387143, 8387358, 8387545, 8387707, 8387847, 8387968, 8388072, 8388161, 8388237, 8388301, 8388355, 8388401, 8388439, 8388471, 8388498, 8388520, 8388538, 8388553, 8388565, 8388575, 8388583, 8388589, 8388594, 8388598, 8388601, 8388603, 8388605, 8388606, 8388607, 8388608};
static const size_t w_statistic_cdf_24[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 903, 1066, 1255, 1472, 1721, 2007, 2333, 2704, 3127, 3606, 4148, 4761, 5451, 6227, 7099, 8075, 9166, 10385, 11742, 13251, 14928, 16785, 18839, 21109, 23611, 26366, 29396, 32721, 36365, 40355, 44715, 49473, 54661, 60306, 66442, 73105, 80327, 88146, 96604, 105737, 115589, 126206, 137629, 149907, 163091, 177227, 192368, 208571, 225886, 244371, 264087, 285088, 307436, 331196, 356425, 383189, 411555, 441583, 473341, 506899, 542318, 579667, 619017, 660429, 703972, 749717, 797723, 848058, 900790, 955976, 1013681, 1073969, 1136892, 1202510, 1270882, 1342054, 1416078, 1493006, 1572875, 1655730, 1741614, 1830553, 1922582, 2017733, 2116021, 2217469, 2322096, 2429904, 2540903, 2655098, 2772478, 2893036, 3016764, 3143634, 3273626, 3406715, 3542857, 3682016, 3824151, 3969202, 4117117, 4267839, 4421292, 4577408, 4736115, 4897321, 5060943, 5226894, 5395068, 5565368, 5737694, 5911926, 6087955, 6265669, 6444937, 6625640, 6807655, 6990843, 7175076, 7360224, 7546141, 7732693, 7919745, 8107147, 8294762, 8482454, 8670069, 8857471, 9044523, 9231075, 9416992, 9602140, 9786373, 9969561, 10151576, 10332279, 10511547, 10689261, 10865290, 11039522, 11211848, 11382148, 11550322, 11716273, 11879895, 12041101, 12199808, 12355924, 12509377, 12660099, 12808014, 12953065, 13095200, 13234359, 13370501, 13503590, 13633582, 13760452, 13884180, 14004738, 14122118, 14236313, 14347312, 14455120, 14559747, 14661195, 14759483, 14854634, 14946663, 15035602, 15121486, 15204341, 15284210, 15361138, 15435162, 15506334, 15574706, 15640324, 15703247, 15763535, 15821240, 15876426, 15929158, 15979493, 16027499, 16073244, 16116787, 16158199, 16197549, 16234898, 16270317, 16303875, 16335633, 16365661, 16394027, 16420791, 16446020, 16469780, 16492128, 16513129, 16532845, 16551330, 16568645, 16584848, 16599989, 16614125, 16627309, 16639587, 16651010, 16661627, 16671479, 16680612, 16689070, 16696889, 16704111, 16710774, 16716910, 16722555, 16727743, 16732501, 16736861, 16740851, 16744495, 16747820, 16750850, 16753605, 16756107, 16758377, 16760431, 16762288, 16763965, 16765474, 16766831, 16768050, 16769141, 16770117, 16770989, 16771765, 16772455, 16773068, 16773610, 16774089, 16774512, 16774883, 16775209, 16775495, 16775744, 16775961, 16776150, 16776313, 16776454, 16776576, 16776680, 16776769, 16776845, 16776909, 16776963, 16777009, 16777047, 16777079, 16777106, 16777128, 16777146, 16777161, 16777173, 16777183, 16777191, 16777197, 16777202, 16777206, 16777209, 16777211, 16777213, 16777214, 16777215, 16777216};
static const size_t w_statistic_cdf_25[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 904, 1068, 1258, 1477, 1728, 2017, 2347, 2723, 3152, 3639, 4191, 4816, 5521, 6315, 7209, 8212, 9335, 10592, 11995, 13558, 15299, 17232, 19375, 21749, 24373, 27269, 30462, 33976, 37837, 42076, 46722, 51806, 57365, 63433, 70048, 77253, 85088, 93597, 102831, 112836, 123664, 135372, 148014, 161649, 176342, 192155, 209153, 227410, 246995, 267982, 290453, 314484, 340157, 367561, 396780, 427904, 461028, 496244, 533647, 573341, 615423, 659994, 707163, 757033, 809709, 865306, 923929, 985687, 1050697, 1119067, 1190908, 1266337, 1345463, 1428396, 1515253, 1606141, 1701166, 1800442, 1904071, 2012155, 2124803, 2242108, 2364165, 2491074, 2622920, 2759787, 2901763, 3048921, 3201332, 3359070, 3522195, 3690759, 3864822, 4044424, 4229602, 4420396, 4616826, 4818908, 5026661, 5240084, 5459171, 5683917, 5914298, 6150283, 6391845, 6638935, 6891496, 7149476, 7412801, 7681389, 7955163, 8234022, 8517859, 8806572, 9100035, 9398118, 9700691, 10007607, 10318710, 10633850, 10952856, 11275550, 11601761, 11931298, 12263964, 12599571, 12937908, 13278763, 13621931, 13967190, 14314313, 14663083, 15013267, 15364629, 15716944, 16069973, 16423473, 16777216, 17130959, 17484459, 17837488, 18189803, 18541165, 18891349, 19240119, 19587242, 19932501, 20275669, 20616524, 20954861, 21290468, 21623134, 21952671, 22278882, 22601576, 22920582, 23235722, 23546825, 23853741, 24156314, 24454397, 24747860, 25036573, 25320410, 25599269, 25873043, 26141631, 26404956, 26662936, 26915497, 27162587, 27404149, 27640134, 27870515, 28095261, 28314348, 28527771, 28735524, 28937606, 29134036, 29324830, 29510008, 29689610, 29863673, 30032237, 30195362, 30353100, 30505511, 30652669, 30794645, 30931512, 31063358, 31190267, 31312324, 31429629, 31542277, 31650361, 31753990, 31853266, 31948291, 32039179, 32126036, 32208969, 32288095, 32363524, 32435365, 32503735, 32568745, 32630503, 32689126, 32744723, 32797399, 32847269, 32894438, 32939009, 32981091, 33020785, 33058188, 33093404, 33126528, 33157652, 33186871, 33214275, 33239948, 33263979, 33286450, 33307437, 33327022, 33345279, 33362277, 33378090, 33392783, 33406418, 33419060, 33430768, 33441596, 33451601, 33460835, 33469344, 33477179, 33484384, 33490999, 33497067, 33502626, 33507710, 33512356, 33516595, 33520456, 33523970, 33527163, 33530059, 33532683, 33535057, 33537200, 33539133, 33540874, 33542437, 33543840, 33545097, 33546220, 33547223, 33548117, 33548911, 33549616, 33550241, 33550793, 33551280, 33551709, 33552085, 33552415, 33552704, 33552955, 33553174, 33553364, 33553528, 33553670, 33553792, 33553896, 33553985, 33554061, 33554125, 33554179, 33554225, 33554263, 33554295, 33554322, 33554344, 33554362, 33554377, 33554389, 33554399, 33554407, 33554413, 33554418, 33554422, 33554425, 33554427, 33554429, 33554430, 33554431, 33554432};
static const size_t w_statistic_cdf_26[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 904, 1069, 1260, 1480, 1733, 2024, 2357, 2737, 3171, 3664, 4224, 4859, 5576, 6385, 7297, 8322, 9472, 10761, 12202, 13811, 15606, 17603, 19822, 22285, 25013, 28031, 31366, 35044, 39095, 43553, 48450, 53823, 59712, 66156, 73200, 80892, 89279, 98413, 108352, 119151, 130873, 143584, 157349, 172241, 188337, 205713, 224452, 244642, 266370, 289731, 314826, 341753, 370619, 401537, 434617, 469980, 507750, 548050, 591012, 636774, 685471, 737247, 792251, 850630, 912540, 978142, 1047593, 1121059, 1198711, 1280716, 1367250, 1458492, 1554616, 1655806, 1762248, 1874123, 1991619, 2114926, 2244228, 2379716, 2521583, 2670012, 2825193, 2987318, 3156567, 3333128, 3517186, 3708915, 3908495, 4116103, 4331904, 4556065, 4788751, 5030111, 5280299, 5539463, 5807734, 6085245, 6372124, 6668480, 6974424, 7290058, 7615464, 7950725, 8295916, 8651090, 9016299, 9391584, 9776966, 10172463, 10578083, 10993809, 11419622, 11855493, 12301367, 12757188, 13222886, 13698366, 14183532, 14678274, 15182458, 15695946, 16218587, 16750206, 17290625, 17839655, 18397079, 18962680, 19536229, 20117473, 20706158, 21302018, 21904763, 22514105, 23129745, 23751362, 24378636, 25011238, 25648818, 26291031, 26937523, 27587921, 28241856, 28898956, 29558829, 30221092, 30885357, 31551219, 32218285, 32886159, 33554432, 34222705, 34890579, 35557645, 36223507, 36887772, 37550035, 38209908, 38867008, 39520943, 40171341, 40817833, 41460046, 42097626, 42730228, 43357502, 43979119, 44594759, 45204101, 45806846, 46402706, 46991391, 47572635, 48146184, 48711785, 49269209, 49818239, 50358658, 50890277, 51412918, 51926406, 52430590, 52925332, 53410498, 53885978, 54351676, 54807497, 55253371, 55689242, 56115055, 56530781, 56936401, 57331898, 57717280, 58092565, 58457774, 58812948, 59158139, 59493400, 59818806, 60134440, 60440384, 60736740, 61023619, 61301130, 61569401, 61828565, 62078753, 62320113, 62552799, 62776960, 62992761, 63200369, 63399949, 63591678, 63775736, 63952297, 64121546, 64283671, 64438852, 64587281, 64729148, 64864636, 64993938, 65117245, 65234741, 65346616, 65453058, 65554248, 65650372, 65741614, 65828148, 65910153, 65987805, 66061271, 66130722, 66196324, 66258234, 66316613, 66371617, 66423393, 66472090, 66517852, 66560814, 66601114, 66638884, 66674247, 66707327, 66738245, 66767111, 66794038, 66819133, 66842494, 66864222, 66884412, 66903151, 66920527, 66936623, 66951515, 66965280, 66977991, 66989713, 67000512, 67010451, 67019585, 67027972, 67035664, 67042708, 67049152, 67055041, 67060414, 67065311, 67069769, 67073820, 67077498, 67080833, 67083851, 67086579, 67089042, 67091261, 67093258, 67095053, 67096662, 67098103, 67099392, 67100542, 67101567, 67102479, 67103288, 67104005, 67104640, 67105200, 67105693, 67106127, 67106507, 67106840, 67107131, 67107384, 67107604, 67107795, 67107960, 67108102, 67108224, 67108328, 67108417, 67108493, 67108557, 67108611, 67108657, 67108695, 67108727, 67108754, 67108776, 67108794, 67108809, 67108821, 67108831, 67108839, 67108845, 67108850, 67108854, 67108857, 67108859, 67108861, 67108862, 67108863, 67108864};
static const size_t w_statistic_cdf_27[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 904, 1069, 1261, 1482, 1736, 2029, 2364, 2747, 3185, 3683, 4249, 4892, 5619, 6440, 7367, 8410, 9582, 10898, 12371, 14018, 15859, 17910, 20193, 22732, 25549, 28671, 32128, 35948, 40164, 44813, 49930, 55556, 61736, 68513, 75937, 84063, 92943, 102637, 113211, 124727, 137258, 150881, 165671, 181713, 199098, 217915, 238263, 260248, 283973, 309553, 337111, 366766, 398650, 432903, 469661, 509075, 551303, 596500, 644835, 696486, 751627, 810447, 873143, 939909, 1010953, 1086494, 1166744, 1251932, 1342295, 1438065, 1539491, 1646829, 1760329, 1880258, 2006890, 2140493, 2281350, 2429752, 2585981, 2750335, 2923120, 3104629, 3295173, 3495068, 3704617, 3924140, 4153960, 4394386, 4645742, 4908354, 5182534, 5468605, 5766893, 6077704, 6401358, 6738174, 7088450, 7452495, 7830616, 8223096, 8630230, 9052306, 9489587, 9942344, 10410842, 10895318, 11396015, 11913167, 12446978, 12997656, 13565401, 14150376, 14752750, 15372679, 16010282, 16665683, 17338989, 18030270, 18739597, 19467025, 20212569, 20976245, 21758050, 22557940, 23375870, 24211779, 25065559, 25937104, 26826287, 27732937, 28656883, 29597934, 30555853, 31530404, 32521329, 33528328, 34551099, 35589321, 36642627, 37710653, 38793016, 39889288, 40999044, 42121842, 43257195, 44404624, 45563631, 46733677, 47914231, 49104746, 50304638, 51513330, 52730234, 53954724, 55186187, 56424001, 57667508, 58916066, 60169026, 61425706, 62685446, 63947578, 65211408, 66476262, 67741466, 69006320, 70270150, 71532282, 72792022, 74048702, 75301662, 76550220, 77793727, 79031541, 80263004, 81487494, 82704398, 83913090, 85112982, 86303497, 87484051, 88654097, 89813104, 90960533, 92095886, 93218684, 94328440, 95424712, 96507075, 97575101, 98628407, 99666629, 100689400, 101696399, 102687324, 103661875, 104619794, 105560845, 106484791, 107391441, 108280624, 109152169, 110005949, 110841858, 111659788, 112459678, 113241483, 114005159, 114750703, 115478131, 116187458, 116878739, 117552045, 118207446, 118845049, 119464978, 120067352, 120652327, 121220072, 121770750, 122304561, 122821713, 123322410, 123806886, 124275384, 124728141, 125165422, 125587498, 125994632, 126387112, 126765233, 127129278, 127479554, 127816370, 128140024, 128450835, 128749123, 129035194, 129309374, 129571986, 129823342, 130063768, 130293588, 130513111, 130722660, 130922555, 131113099, 131294608, 131467393, 131631747, 131787976, 131936378, 132077235, 132210838, 132337470, 132457399, 132570899, 132678237, 132779663, 132875433, 132965796, 133050984, 133131234, 133206775, 133277819, 133344585, 133407281, 133466101, 133521242, 133572893, 133621228, 133666425, 133708653, 133748067, 133784825, 133819078, 133850962, 133880617, 133908175, 133933755, 133957480, 133979465, 133999813, 134018630, 134036015, 134052057, 134066847, 134080470, 134093001, 134104517, 134115091, 134124785, 134133665, 134141791, 134149215, 134155992, 134162172, 134167798, 134172915, 134177564, 134181780, 134185600, 134189057, 134192179, 134194996, 134197535, 134199818, 134201869, 134203710, 134205357, 134206830, 134208146, 134209318, 134210361, 134211288, 134212109, 134212836, 134213479, 134214045, 134214543, 134214981, 134215364, 134215699, 134215992, 134216246, 134216467, 134216659, 134216824, 134216966, 134217088, 134217192, 134217281, 134217357, 134217421, 134217475, 134217521, 134217559, 134217591, 134217618, 134217640, 134217658, 134217673, 134217685, 134217695, 134217703, 134217709, 134217714, 134217718, 134217721, 134217723, 134217725, 134217726, 134217727, 134217728};
static const size_t w_statistic_cdf_28[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 904, 1069, 1261, 1483, 1738, 2032, 2369, 2754, 3195, 3697, 4268, 4917, 5652, 6483, 7422, 8480, 9670, 11008, 12508, 14187, 16066, 18163, 20500, 23103, 25996, 29207, 32768, 36710, 41068, 45882, 51191, 57038, 63472, 70542, 78301, 86810, 96128, 106320, 117460, 129619, 142877, 157321, 173038, 190123, 208680, 228813, 250634, 274266, 299832, 327463, 357304, 389498, 424199, 461574, 501789, 545023, 591467, 641313, 694765, 752042, 813363, 878960, 949080, 1023972, 1103896, 1189131, 1279955, 1376659, 1479553, 1588946, 1705162, 1828542, 1959427, 2098173, 2245153, 2400741, 2565323, 2739305, 2923092, 3117101, 3321770, 3537532, 3764834, 4004143, 4255920, 4520640, 4798795, 5090872, 5397369, 5718801, 6055677, 6408514, 6777846, 7164198, 7568102, 7990106, 8430745, 8890560, 9370107, 9869925, 10390559, 10932564, 11496477, 12082837, 12692192, 13325070, 13981996, 14663502, 15370098, 16102285, 16860574, 17645444, 18457367, 19296819, 20164242, 21060069, 21984731, 22938624, 23922131, 24935630, 25979462, 27053949, 28159408, 29296114, 30464320, 31664274, 32896175, 34160200, 35456517, 36785243, 38146470, 39540278, 40966695, 42425722, 43917344, 45441495, 46998077, 48586977, 50208028, 51861029, 53545766, 55261967, 57009326, 58787525, 60596184, 62434894, 64303228, 66200702, 68126800, 70080991, 72062688, 74071270, 76106104, 78166503, 80251746, 82361105, 84493795, 86649003, 88825909, 91023640, 93241299, 95477982, 97732737, 100004590, 102292565, 104595641, 106912777, 109242935, 111585038, 113937990, 116300706, 118672062, 121050922, 123436165, 125826635, 128221171, 130618629, 133017836, 135417620, 137816827, 140214285, 142608821, 144999291, 147384534, 149763394, 152134750, 154497466, 156850418, 159192521, 161522679, 163839815, 166142891, 168430866, 170702719, 172957474, 175194157, 177411816, 179609547, 181786453, 183941661, 186074351, 188183710, 190268953, 192329352, 194364186, 196372768, 198354465, 200308656, 202234754, 204132228, 206000562, 207839272, 209647931, 211426130, 213173489, 214889690, 216574427, 218227428, 219848479, 221437379, 222993961, 224518112, 226009734, 227468761, 228895178, 230288986, 231650213, 232978939, 234275256, 235539281, 236771182, 237971136, 239139342, 240276048, 241381507, 242455994, 243499826, 244513325, 245496832, 246450725, 247375387, 248271214, 249138637, 249978089, 250790012, 251574882, 252333171, 253065358, 253771954, 254453460, 255110386, 255743264, 256352619, 256938979, 257502892, 258044897, 258565531, 259065349, 259544896, 260004711, 260445350, 260867354, 261271258, 261657610, 262026942, 262379779, 262716655, 263038087, 263344584, 263636661, 263914816, 264179536, 264431313, 264670622, 264897924, 265113686, 265318355, 265512364, 265696151, 265870133, 266034715, 266190303, 266337283, 266476029, 266606914, 266730294, 266846510, 266955903, 267058797, 267155501, 267246325, 267331560, 267411484, 267486376, 267556496, 267622093, 267683414, 267740691, 267794143, 267843989, 267890433, 267933667, 267973882, 268011257, 268045958, 268078152, 268107993, 268135624, 268161190, 268184822, 268206643, 268226776, 268245333, 268262418, 268278135, 268292579, 268305837, 268317996, 268329136, 268339328, 268348646, 268357155, 268364914, 268371984, 268378418, 268384265, 268389574, 268394388, 268398746, 268402688, 268406249, 268409460, 268412353, 268414956, 268417293, 268419390, 268421269, 268422948, 268424448, 268425786, 268426976, 268428034, 268428973, 268429804, 268430539, 268431188, 268431759, 268432261, 268432702, 268433087, 268433424, 268433718, 268433973, 268434195, 268434387, 268434552, 268434694, 268434816, 268434920, 268435009, 268435085, 268435149, 268435203, 268435249, 268435287, 268435319, 268435346, 268435368, 268435386, 268435401, 268435413, 268435423, 268435431, 268435437, 268435442, 268435446, 268435449, 268435451, 268435453, 268435454, 268435455, 268435456};
static const size_t w_statistic_cdf_29[] = {
1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 904, 1069, 1261, 1483, 1739, 2034, 2372, 2759, 3202, 3707, 4282, 4936, 5677, 6516, 7465, 8535, 9740, 11096, 12618, 14324, 16235, 18370, 20753, 23410, 26367, 29654, 33304, 37350, 41830, 46786, 52260, 58299, 64955, 72280, 80333, 89179, 98882, 109515, 121157, 133887, 147794, 162973, 179521, 197545, 217160, 238483, 261642, 286774, 314019, 343529, 375467, 409998, 447302, 487570, 530996, 577791, 628177, 682381, 740647, 803233, 870401, 942432, 1019622, 1102273, 1190706, 1285259, 1386275, 1494119, 1609172, 1731823, 1862483, 2001580, 2149550, 2306853, 2473966, 2651375, 2839589, 3039137, 3250555, 3474405, 3711268, 3961731, 4226408, 4505932, 4800943, 5112107, 5440108, 5785637, 6149411, 6532164, 6934637, 7357594, 7801818, 8268094, 8757233, 9270061, 9807404, 10370113, 10959053, 11575087, 12219101, 12891991, 13594650, 14327990, 15092933, 15890393, 16721301, 17586594, 18487199, 19424055, 20398106, 21410278, 22461510, 23552739, 24684882, 25858864, 27075603, 28335993, 29640932, 30991307, 32387976, 33831795, 35323606, 36864216, 38454426, 40095019, 41786735, 43530307, 45326442, 47175802, 49079034, 51036755, 53049532, 55117914, 57242414, 59423491, 61661579, 63957075, 66310313, 68721603, 71191210, 73719334, 76306145, 78951767, 81656253, 84419625, 87241852, 90122833, 93062430, 96060453, 99116637, 102230678, 105402218, 108630823, 111916020, 115257280, 118653995, 122105520, 125611152, 129170110, 132781577, 136444677, 140158459, 143921934, 147734060, 151593718, 155499754, 159450963, 163446067, 167483756, 171562673, 175681388, 179838447, 184032349, 188261529, 192524399, 196819331, 201144636, 205498611, 209879515, 214285555, 218714925, 223165794, 227636280, 232124499, 236628545, 241146469, 245676327, 250216161, 254763978, 259317797, 263875628, 268435456, 272995284, 277553115, 282106934, 286654751, 291194585, 295724443, 300242367, 304746413, 309234632, 313705118, 318155987, 322585357, 326991397, 331372301, 335726276, 340051581, 344346513, 348609383, 352838563, 357032465, 361189524, 365308239, 369387156, 373424845, 377419949, 381371158, 385277194, 389136852, 392948978, 396712453, 400426235, 404089335, 407700802, 411259760, 414765392, 418216917, 421613632, 424954892, 428240089, 431468694, 434640234, 437754275, 440810459, 443808482, 446748079, 449629060, 452451287, 455214659, 457919145, 460564767, 463151578, 465679702, 468149309, 470560599, 472913837, 475209333, 477447421, 479628498, 481752998, 483821380, 485834157, 487791878, 489695110, 491544470, 493340605, 495084177, 496775893, 498416486, 500006696, 501547306, 503039117, 504482936, 505879605, 507229980, 508534919, 509795309, 511012048, 512186030, 513318173, 514409402, 515460634, 516472806, 517446857, 518383713, 519284318, 520149611, 520980519, 521777979, 522542922, 523276262, 523978921, 524651811, 525295825, 525911859, 526500799, 527063508, 527600851, 528113679, 528602818, 529069094, 529513318, 529936275, 530338748, 530721501, 531085275, 531430804, 531758805, 532069969, 532364980, 532644504, 532909181, 533159644, 533396507, 533620357, 533831775, 534031323, 534219537, 534396946, 534564059, 534721362, 534869332, 535008429, 535139089, 535261740, 535376793, 535484637, 535585653, 535680206, 535768639, 535851290, 535928480, 536000511, 536067679, 536130265, 536188531, 536242735, 536293121, 536339916, 536383342, 536423610, 536460914, 536495445, 536527383, 536556893, 536584138, 536609270, 536632429, 536653752, 536673367, 536691391, 536707939, 536723118, 536737025, 536749755, 536761397, 536772030, 536781733, 536790579, 536798632, 536805957, 536812613, 536818652, 536824126, 536829082, 536833562, 536837608, 536841258, 536844545, 536847502, 536850159, 536852542, 536854677, 536856588, 536858294, 536859816, 536861172, 536862377, 536863447, 536864396, 536865235, 536865976, 536866630, 536867205, 536867710, 536868153, 536868540, 536868878, 536869173, 536869429, 536869651, 536869843, 536870008, 536870150, 536870272, 536870376, 536870465, 536870541, 536870605, 536870659, 536870705, 536870743, 536870775, 536870802, 536870824, 536870842, 536870857, 536870869, 536870879, 536870887, 536870893, 536870898, 536870902, 536870905, 536870907, 536870909, 536870910, 536870911, 536870912};
static const size_t w_statistic_cdf_30[] = {
    1, 2, 3, 5, 7, 10, 14, 19, 25, 33, 43, 55, 70, 88, 110, 137, 169, 207, 253, 307, 371, 447, 536, 640, 762, 904, 1069, 1261, 1483, 1739, 2035, 2374, 2762, 3207, 3714, 4292, 4950, 5696, 6541, 7498, 8578, 9795, 11166, 12706, 14434, 16372, 18539, 20960, 23663, 26674, 30025, 33751, 37886, 42470, 47548, 53164, 59368, 66216, 73763, 82072, 91213, 101254, 112274, 124359, 137594, 152076, 167909, 185198, 204061, 224625, 247018, 271382, 297870, 326637, 357853, 391702, 428368, 468055, 510980, 557363, 607445, 661481, 719731, 782477, 850019, 922661, 1000731, 1084577, 1174553, 1271039, 1374438, 1485157, 1603634, 1730329, 1865710, 2010277, 2164553, 2329071, 2504398, 2691126, 2889858, 3101231, 3325911, 3564574, 3817934, 4086735, 4371729, 4673710, 4993502, 5331939, 5689898, 6068285, 6468018, 6890058, 7335397, 7805038, 8300026, 8821440, 9370367, 9947939, 10555320, 11193679, 11864232, 12568225, 13306910, 14081584, 14893571, 15744200, 16634843, 17566899, 18541768, 19560890, 20625731, 21737754, 22898460, 24109374, 25372009, 26687918, 28058671, 29485825, 30970971, 32515711, 34121630, 35790343, 37523471, 39322613, 41189389, 43125424, 45132310, 47211659, 49365080, 51594139, 53900420, 56285495, 58750889, 61298135, 63928746, 66644182, 69445904, 72335347, 75313884, 78382880, 81543669, 84797512, 88145658, 91589316, 95129612, 98767655, 102504506, 106341135, 110278489, 114317455, 118458826, 122703362, 127051760, 131504613, 136062473, 140725824, 145495039, 150370446, 155352299, 160440730, 165635827, 170937594, 176345912, 181860611, 187481432, 193207991, 199039848, 204976474, 211017209, 217161333, 223408038, 229756380, 236205359, 242753883, 249400722, 256144592, 262984116, 269917782, 276944024, 284061183, 291267469, 298561041, 305939968, 313402192, 320945603, 328568012, 336267103, 344040519, 351885825, 359800464, 367781847, 375827313, 383934088, 392099374, 400320305, 408593915, 416917218, 425287175, 433700652, 442154505, 450645548, 459170510, 467726123, 476309086, 484916020, 493543565, 502188336, 510846886, 519515796, 528191632, 536870912, 545550192, 554226028, 562894938, 571553488, 580198259, 588825804, 597432738, 606015701, 614571314, 623096276, 631587319, 640041172, 648454649, 656824606, 665147909, 673421519, 681642450, 689807736, 697914511, 705959977, 713941360, 721855999, 729701305, 737474721, 745173812, 752796221, 760339632, 767801856, 775180783, 782474355, 789680641, 796797800, 803824042, 810757708, 817597232, 824341102, 830987941, 837536465, 843985444, 850333786, 856580491, 862724615, 868765350, 874701976, 880533833, 886260392, 891881213, 897395912, 902804230, 908105997, 913301094, 918389525, 923371378, 928246785, 933016000, 937679351, 942237211, 946690064, 951038462, 955282998, 959424369, 963463335, 967400689, 971237318, 974974169, 978612212, 982152508, 985596166, 988944312, 992198155, 995358944, 998427940, 1001406477, 1004295920, 1007097642, 1009813078, 1012443689, 1014990935, 1017456329, 1019841404, 1022147685, 1024376744, 1026530165, 1028609514, 1030616400, 1032552435, 1034419211, 1036218353, 1037951481, 1039620194, 1041226113, 1042770853, 1044255999, 1045683153, 1047053906, 1048369815, 1049632450, 1050843364, 1052004070, 1053116093, 1054180934, 1055200056, 1056174925, 1057106981, 1057997624, 1058848253, 1059660240, 1060434914, 1061173599, 1061877592, 1062548145, 1063186504, 1063793885, 1064371457, 1064920384, 1065441798, 1065936786, 1066406427, 1066851766, 1067273806, 1067673539, 1068051926, 1068409885, 1068748322, 1069068114, 1069370095, 1069655089, 1069923890, 1070177250, 1070415913, 1070640593, 1070851966, 1071050698, 1071237426, 1071412753, 1071577271, 1071731547, 1071876114, 1072011495, 1072138190, 1072256667, 1072367386, 1072470785, 1072567271, 1072657247, 1072741093, 1072819163, 1072891805, 1072959347, 1073022093, 1073080343, 1073134379, 1073184461, 1073230844, 1073273769, 1073313456, 1073350122, 1073383971, 1073415187, 1073443954, 1073470442, 1073494806, 1073517199, 1073537763, 1073556626, 1073573915, 1073589748, 1073604230, 1073617465, 1073629550, 1073640570, 1073650611, 1073659752, 1073668061, 1073675608, 1073682456, 1073688660, 1073694276, 1073699354, 1073703938, 1073708073, 1073711799, 1073715150, 1073718161, 1073720864, 1073723285, 1073725452, 1073727390, 1073729118, 1073730658, 1073732029, 1073733246, 1073734326, 1073735283, 1073736128, 1073736874, 1073737532, 1073738110, 1073738617, 1073739062, 1073739450, 1073739789, 1073740085, 1073740341, 1073740563, 1073740755, 1073740920, 1073741062, 1073741184, 1073741288, 1073741377, 1073741453, 1073741517, 1073741571, 1073741617, 1073741655, 1073741687, 1073741714, 1073741736, 1073741754, 1073741769, 1073741781, 1073741791, 1073741799, 1073741805, 1073741810, 1073741814, 1073741817, 1073741819, 1073741821, 1073741822, 1073741823, 1073741824};

const static size_t* pointer_array[] = {
    NULL, w_statistic_cdf_1, w_statistic_cdf_2, w_statistic_cdf_3,
    w_statistic_cdf_4, w_statistic_cdf_5, w_statistic_cdf_6,    
    w_statistic_cdf_7, w_statistic_cdf_8, w_statistic_cdf_9,    
    w_statistic_cdf_10, w_statistic_cdf_11, w_statistic_cdf_12,    
    w_statistic_cdf_13, w_statistic_cdf_14, w_statistic_cdf_15,    
    w_statistic_cdf_16, w_statistic_cdf_17, w_statistic_cdf_18,    
    w_statistic_cdf_19, w_statistic_cdf_20, w_statistic_cdf_21,    
    w_statistic_cdf_22, w_statistic_cdf_23, w_statistic_cdf_24,    
    w_statistic_cdf_25, w_statistic_cdf_26, w_statistic_cdf_27,    
    w_statistic_cdf_28, w_statistic_cdf_29, w_statistic_cdf_30};

const static vector<const size_t*> pointer_vec(pointer_array, pointer_array + 31);

static double
wilcoxon_p_value(const double &w_pos,
                 const double &w_neg,
                 const size_t &sample_size,
                 const bool &alternative)
{
    if (sample_size <= 30) // exact value
    {
        const size_t w_max = sample_size * (sample_size + 1) / 2;
        assert(w_pos <= w_max && w_neg <= w_max);
        const size_t* ptr = pointer_vec[sample_size];
        if (alternative)
        {
            const double smaller =
                *(ptr + static_cast<size_t>(floor(w_neg == 0 ? 0 : w_neg - 1)));
            return 1.0 - smaller / *(ptr + w_max);
        }
        else
        {
            const double smaller =
                *(ptr + static_cast<size_t>(floor(w_pos == 0 ? 0 : w_pos - 1)));
            return 1.0 - smaller / *(ptr + w_max);
        }
    }
    else // normal approximation
    {
        const size_t w_max = sample_size * (sample_size + 1) / 2;
        assert(w_pos <= w_max && w_neg <= w_max);

        const double mu = sample_size * (sample_size + 1) / 4.0;
        const double var =
            sample_size * (sample_size + 1) * (2 * sample_size + 1) / 24.0;
        
        if (alternative)
        {
            const double z_neg = (w_neg - mu) / sqrt(var);
            return
                gsl_cdf_gaussian_Q(z_neg, 1.0);
        }
        else
        {
            const double z_pos = (w_pos - mu) / sqrt(var);
            return
                gsl_cdf_gaussian_Q(z_pos, 1.0);
        }
    }
}
